#!/usr/bin/perl
#Rene Warren, April 2008

#cluster headnode name;headnode;apollo
#cluster request resource;resource;-l tmp_free=2G
#qsub command;qsub;/opt/sge/bin/lx24-amd64/qsub
#qstat command;qstat;/opt/sge/bin/lx24-amd64/qstat
#wait for n sec. before running qstat;wait;120
#location of sequence database file -filter out-;seqcluster;/projects/02/rwarren/solexa/immunoprofiling/simulation/SSAKE/parallelSSAKE/POOL-V.fa
#Number of jobs to run on cluster 1st batch;cluster_jobs;100
#location of assembler;assembler;/projects/02/rwarren/solexa/immunoprofiling/simulation/SSAKE/SSAKErr
#assembler options;options;-m 20 -o 2 -r 0.6                   


use strict;
use Data::Dumper;

my $config = "assembly.config";
my $pipeline = "na";
my $mode = "serial";


if($#ARGV<2){die "Usage: $0 <query fasta e.g. SEEDS.fa.normalized.fa> <base name for files> <serial/parallel> <pipeline name (optional) -clean,finish- >\n";}
if(! -e $config){die "Error: can't find $config, make sure it is in your working directory -- fatal\n";}
if(! -e $ARGV[0]){die "Error: can't find file $ARGV[0] -- fatal\n";}

### Getting configuration information
my $conf;
open(CONFIG,$config);
while(<CONFIG>){
   chomp;
   my @a=split(/\;/);
   $conf->{$a[1]}=$a[2];
}
close CONFIG;

$pipeline = $ARGV[3] if ($ARGV[3] ne "");
$mode = "parallel" if ($ARGV[2] ne "serial");
my $dir = `pwd`;
chomp($dir);
my $base_job_name = $ARGV[1];
my $TIME_TO_WAIT = $conf->{'wait'};
my $log = $dir . "/qstat.txt";

if($pipeline eq "clean"){
   removeDataFromNodes($dir, $conf->{'headnode'});
   exit;  
}


### Launching the first job
my $base_name = "SDB1";
my $screen = "parallel100";
my $super_base_name = $base_job_name . $base_name;
my $jobdir = $dir . "/" . $screen;


eval{


if($pipeline ne "finish"){

   ###----------------------------------
   ### Split Original Input
   `rm -rf $jobdir`;
   `mkdir $jobdir`;
   my $first_file_input = &splitInput($ARGV[0],$jobdir,$conf->{'cluster_jobs'});

   my $db = $conf->{'db'};
   if(! -e $db){die "*$db* doesn't exist..check your input.\n";}
   my $db_name = $1 if ($db=~/([^\/]*)$/);

   print "Generating Shell Scripts...\n";
   my $job_files=&makeSS($conf->{'assembler'}, $conf->{'options'}, $first_file_input,$jobdir, $super_base_name,$base_name,$db_name,$db,$mode);

   ### Execute jobs
   if($mode eq "parallel"){
      &executeJobs($job_files, $jobdir, $super_base_name, $conf->{'headnode'}, $conf->{'qsub'}, $conf->{'resource'});
      ### Monitor/wait for jobs to finish
      &monitorJobs($super_base_name, $log, $conf->{'headnode'}, $conf->{'qstat'});
   }else{
      foreach my $runfile (@$job_files){
         `chmod 755 $runfile`;
         my $cmd = "bash $runfile";
         print "Will Execute: $cmd\n";
         system($cmd);
      } 
   }
}
#--------------------concat final hits into a single file


};###end eval block


my $date = `date`;

if($@){
   my $message = $@;
   my $failure = "\nSomething went wrong running $0 $date\n$message\n";
   print $failure;
   print LOG $failure;
}else{
   my $success = "\nScript executed normally $date\n";
   print $success;
   print LOG $success;
}

exit;








#------------------------------------------------
sub removeDataFromNodes{

   my ($dir, $headnode, $qsub) = @_;
   
   my $jobdir = $dir . "/clean";
   `rm -rf $jobdir`;
   `mkdir $jobdir`;
   
   for(my $x=1;$x<=150;$x++){
      my $file = "cleanNodes" . $x . ".sh";
      open(CLEAN, ">$jobdir/$file");
      print CLEAN "#!/bin/sh\n";
      print CLEAN "#\$ -S /bin/sh\n";
      print CLEAN "#\$ -V\n";
      print CLEAN "rm -rf /tmp/*";
      close CLEAN; 
   }

   for(my $x=1;$x<=150;$x++){
      my $file = $jobdir . "/cleanNodes" . $x . ".sh";
      `chmod 755 $file`;
      my $qsub_cmd= "ssh $headnode 'export SGE_ROOT=/opt/sge;cd $jobdir;$qsub $file' ";
      system($qsub_cmd);
   }
}

#------------------------------------------------
sub executeJobs{

   my ($job_file, $jobdir, $super_base_name, $headnode, $qsub, $resource) = @_;

   my $job_count=0;

   foreach my $job_file(@$job_file){
      $job_count++;
      my $subjobname=$super_base_name . "." . $job_count;
      my $qsub_cmd= "ssh $headnode 'export SGE_ROOT=/opt/sge;cd $jobdir;$qsub $resource -N $subjobname $job_file' "; 
      print "$qsub_cmd\n";
      system($qsub_cmd);
      sleep(1);
   }
}

#------------------------------------------------
#Check status, wait until all sub jobs are done before resuming script
#The system call returns 0 or 1 depending on success or failure, respectively (yes 0==success!)
#I am not aware of other ways to capture the output generated by programs called by system, but to redirect the output to a file.
sub monitorJobs{

   my ($super_base_name, $log, $headnode, $qstat) = @_;

   my $flag=1;
   my @status;

   while($flag){

      print "There are still some jobs queued/running...will wait $TIME_TO_WAIT sec. jobs are $super_base_name*\n";
      sleep($TIME_TO_WAIT);
      my $qstat_cmd="ssh $headnode 'export SGE_ROOT=/opt/sge;$qstat > $log'";
      system($qstat_cmd);

      #`cat $log1 $log2 > $log`;
      #sleep(2);
      open (LOG, $log) || die "Can't open $log.\n";
      @status=<LOG>;
      close LOG;

      $flag=0; #reset flag
      foreach my $status_line(@status){
          if($status_line=~/$super_base_name/i){  ###this will capture only jobs related to this project
            $flag=1;
            print "$status_line\n";
         }
      }
   }

   print "All sub jobs of run $super_base_name have finished.\n";
}

#------------------------------------------
sub makeSS{

   my ($exec, $options, $file_input,$dir,$super_base_name,$base_name,$db_name,$db,$mode)=@_;   

   my @job_files;
   
   my $job_id=1;
   foreach my $input (@$file_input){

      my $output;
      my $tmp_output;  
 
      my $title;
      chomp($input);      

         my $input_fn;

         if ($input=~/([^\/]*)$/){
            my $filename=$1;
            $output = $dir ."/". $super_base_name . "." . $job_id . "_" . $db_name .".out";
            $tmp_output = "/tmp/" . $super_base_name . "." . $job_id . "_" . $db_name .".out";            

            $title = $filename . "_" . $db_name;
            $input_fn = $filename;
         }

         my $work_dir = "/tmp/" . $base_name . $job_id;
         my $work_all = $work_dir . "/*";
         my $new_in = $work_dir . "/" . $input_fn; 
         my $new_db = $work_dir . "/" . $db_name;
         
         my $job_file=$dir . "/" . $super_base_name . "." . $job_id . ".sh";

         push @job_files, $job_file;
         open (JOB, ">$job_file") or die "can't open $job_file for writing.\n";
         print JOB "#!/bin/sh\n";
         print JOB "#\$ -S /bin/sh\n";
         print JOB "#\$ -V\n";
         my $cmd1 = "rm -rf $work_all;rm -rf $work_dir;mkdir $work_dir;cp -rf $db $work_dir/.;cp -rf $input $new_in;cd $dir;$exec -s $new_in -f $new_db $options $job_id;cp -rf $work_dir/*ssake* $dir/.;rm -rf $work_all;rm -rf $work_dir\n";
         my $cmd2 = "$exec -f $db -s $input $options\n";
         print JOB $cmd1 if($mode eq 'parallel');
         print JOB $cmd2 if($mode eq 'serial');
         close JOB;
         $job_id++;
   }

   return \@job_files;
}

#--------------------------------------------
sub splitInput{

   my ($fa,$dir,$numjobs) = @_;

   my @file_input;

   ### Defining the number of entries in each job file
   my $count = `grep -c ">" $fa`;
   chomp($count);
   my $entries = int($count / $numjobs);

   ### splitting input fasta file
   print "There are $count entries in your input fasta file $fa.  Will split into $numjobs files each having ~$entries fasta entries.\n";

   open (OUT,$fa) || die "Can't open handle $fa for reading -- fatal.\n";;

   my ($ct,$mod)=(0,1);

   my $filename = $1 if ($fa =~ /\/([^\/]*)$/);

   my $modfa = $filename . "." . $mod;
   my $splitfasta = $dir . "/" . $modfa;
   push @file_input, $splitfasta;

   open (IN, ">$splitfasta") || die "Can't open handle $splitfasta for writing -- fatal.\n";

   my $head="";
   my $unique;
   my $lcr=0;
   my $dup = 0;
   my $tot = 0;

   while(<OUT>){
      chomp;
      if(/(\>\S+)/){
         $head=$1;
      }elsif(/([ACGTNX]*)/i){
         my $seq = $1;

            $ct++;
            print IN "$head\n$seq\n" if ($ct <= $entries);

            if($ct > $entries){
               $tot+=$ct;
               $ct = 0;
               $mod++;
               $modfa = $filename . "." . $mod;
               close IN;
               $splitfasta = $dir . "/" . $modfa;
               open (IN, ">$splitfasta") || die "Can't open handle $splitfasta for writing -- fatal.\n";
               push @file_input, $splitfasta;
            }
      }
   }
   $unique={};
   $tot+=$ct;

   close OUT;
   close IN;

   print "Total in split files: $tot\n";

   return \@file_input;
}
